﻿//-----------------------------------------------------------------------
// <copyright file="Program.cs" company="SVNStyleCop">
//     Copyright (c) SVNStyleCop. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------
namespace SVNStyleCop
{
    using System;
    using System.Collections.Generic;
    using System.Configuration;
    using System.Diagnostics;
    using System.IO;
    using System.Reflection;
    using System.Text;
    using System.Text.RegularExpressions;
    using CoreLibrary;
    using StyleCop;

    /// <summary>
    /// Program main class. Implements the SVNStyleCop engine.
    /// </summary>
    public class Program
    {
        /// <summary>
        /// Configuration section "svnStyleCopConfig" from the App.config file.
        /// </summary>
        private static SvnStyleCopConfigSection configSection;

        /// <summary>
        /// Unique ID of the current instance. Used to create a unique temporary directory.
        /// </summary>
        private static Guid instanceId;

        /// <summary>
        /// All config files found while traversing through the repository.
        /// Used as a cache, so we don't have to search the config for every file.
        /// </summary>
        private static Dictionary<string, string> styleCopConfigs = new Dictionary<string, string>(StringComparer.InvariantCultureIgnoreCase);

        /// <summary>
        /// Main method.
        /// </summary>
        /// <param name="args">Command line arguments</param>
        /// <returns>
        /// 0 if successful, negative values for general errors; positive values indicate the number of violations found
        /// </returns>
        public static int Main(string[] args)
        {
            CommandLineArgument cla = new CommandLineArgument(args);
            if (!cla.Exists("repository") || !(cla.Exists("revision") || cla.Exists("transaction")))
            {
                PrintUsage();
                return -1;
            }

            configSection = (SvnStyleCopConfigSection)ConfigurationManager.GetSection("svnStyleCopConfig");
            instanceId = Guid.NewGuid();

            Dictionary<string, List<string>> localSourceFiles = new Dictionary<string, List<string>>();

            SaveCommittedFilesToTemp(localSourceFiles, cla["repository"], cla["revision"], cla["transaction"]);

            string settingsFileOverride = string.Empty;
            if (cla.Exists("settings"))
            {
                settingsFileOverride = cla["settings"];
            }

            List<Violation> violations = ValidateCommittedFiles(localSourceFiles, settingsFileOverride);

            DeleteTempFoler();

            if (violations.Count > configSection.StyleCop.MaxViolationCount)
            {
                Console.Error.WriteLine("Only first {0} of total {1} violations reported below", configSection.StyleCop.MaxViolationCount, violations.Count);
            }
            else
            {
                Console.Error.WriteLine("{0} violations found", violations.Count);
            }

            for (int i = 0; i < (violations.Count > configSection.StyleCop.MaxViolationCount ? configSection.StyleCop.MaxViolationCount : violations.Count); i++)
            {
                Violation violation = violations[i];
                Console.Error.WriteLine("{0}: {1} in {3}:{2}", violation.Rule.CheckId, violation.Message, violation.Line, violation.SourceCode.Name);
            }

            return violations.Count;
        }

        /// <summary>
        /// Gets all files of the current transaction from the repository and stores them in the temporary directory.
        /// </summary>
        /// <param name="localSourceFiles">Output: mapping of StyleCop config file to files to be checked with that file.</param>
        /// <param name="repositoryPath">path of the repository on the local disk.</param>
        /// <param name="revision">revision number provided by SVN</param>
        /// <param name="transaction">transaction number provided by SVN</param>
        private static void SaveCommittedFilesToTemp(Dictionary<string, List<string>> localSourceFiles, string repositoryPath, string revision, string transaction)
        {
            ProcessStartInfo svnLookPsi = new ProcessStartInfo(configSection.SvnLook.Location);
            svnLookPsi.UseShellExecute = svnLookPsi.ErrorDialog = false;
            svnLookPsi.RedirectStandardOutput = true;
            Process svnLookProcess = new Process();
            svnLookProcess.StartInfo = svnLookPsi;

            string revOrTran = !string.IsNullOrEmpty(revision) ? string.Format("-r {0}", revision) : string.Format("-t {0}", transaction);
            svnLookPsi.Arguments = string.Format("changed \"{0}\" {1}", repositoryPath, revOrTran);

            svnLookProcess.Start();

            string[] committedFiles = svnLookProcess.StandardOutput.ReadToEnd().Split(new[] { "\r\n" }, StringSplitOptions.RemoveEmptyEntries);
            Regex fileNameExtractor = new Regex("^[AU] *(.*)");
            string tempFolder = GetTempFolderName();
            
            foreach (string committedRow in committedFiles)
            {
                string committedFile = fileNameExtractor.Match(committedRow).Groups[1].Value;
                if (NeedToAnalyzeFile(committedFile))
                {
                    string outFileName = Path.Combine(tempFolder, committedFile);
                    Directory.CreateDirectory(Path.GetDirectoryName(outFileName));

                    svnLookPsi.Arguments = string.Format("cat \"{0}\" \"{1}\" {2}", repositoryPath, committedFile, revOrTran);

                    svnLookProcess.Start();

                    StreamWriter outFile = new StreamWriter(outFileName, false, Encoding.UTF8);
                    outFile.Write(svnLookProcess.StandardOutput.ReadToEnd());
                    outFile.Close();

                    string settings = FindStyleCopSettings(committedFile, repositoryPath, revOrTran);
                    List<string> files;
                    if (!localSourceFiles.TryGetValue(settings, out files))
                    {
                        files = new List<string>();
                        localSourceFiles.Add(settings, files);
                    }

                    files.Add(outFileName);
                }
            }
        }

        /// <summary>
        /// Traverses through the repository and searches for a StyleCop.Settings file.
        /// </summary>
        /// <param name="committedFile">File to search the settings for</param>
        /// <param name="repositoryPath">path of the repository on the local disk.</param>
        /// <param name="revOrTran">argument to SvnLook cat command</param>
        /// <returns>the name of the local temporary file containing the settings or an empty string if no settings were found</returns>
        private static string FindStyleCopSettings(string committedFile, string repositoryPath, string revOrTran)
        {
            string filePath = Path.GetDirectoryName(committedFile);

            ProcessStartInfo svnLookPsi = new ProcessStartInfo(configSection.SvnLook.Location);
            svnLookPsi.UseShellExecute = svnLookPsi.ErrorDialog = false;
            svnLookPsi.RedirectStandardOutput = true;
            svnLookPsi.RedirectStandardError = true;
            Process svnLookProcess = new Process();
            svnLookProcess.StartInfo = svnLookPsi;

            while (filePath != null && filePath.Length > 1)
            {
                string settingsFile = filePath.Replace('\\', '/') + "/" + Settings.DefaultFileName;
                string foundSettingsFile;
                if (styleCopConfigs.TryGetValue(settingsFile, out foundSettingsFile))
                {
                    return foundSettingsFile;
                }

                svnLookPsi.Arguments = string.Format(
                    "cat \"{0}\" \"{1}\" {2}", repositoryPath, settingsFile, revOrTran);

                svnLookProcess.Start();
                svnLookProcess.WaitForExit();

                if (svnLookProcess.ExitCode == 0)
                {
                    ////Console.Error.WriteLine("Found config {0} for {1}", settingsFile, committedFile);
                    foundSettingsFile = Path.Combine(GetTempFolderName(), Guid.NewGuid() + ".Settings");
                    StreamWriter outFile = new StreamWriter(foundSettingsFile, false, Encoding.UTF8);
                    outFile.Write(svnLookProcess.StandardOutput.ReadToEnd());
                    outFile.Close();
                    styleCopConfigs.Add(settingsFile, foundSettingsFile);
                    return foundSettingsFile;
                }
                else
                {
                    svnLookProcess.StandardError.ReadToEnd();
                }

                filePath = Path.GetDirectoryName(filePath);
            }

            return string.Empty;
        }

        /// <summary>
        /// Checks with the PathPatterns settings whether the given file should be analyzed or ignored.
        /// </summary>
        /// <param name="filePath">SVN path (inside the repository) of the file</param>
        /// <returns>true iff we have to analyze the file.</returns>
        private static bool NeedToAnalyzeFile(string filePath)
        {
            bool matches = false;
            foreach (PathPatternElement pathPattern in configSection.PathPatterns)
            {
                if (pathPattern.ValueRegex.IsMatch(filePath))
                {
                    if (pathPattern.Ignore)
                    {
                        return false;
                    }

                    matches = true;
                }
            }

            return matches;
        }

        /// <summary>
        /// Validates all files with StyleCop.
        /// </summary>
        /// <param name="localSourceFiles">mapping of StyleCop config file to files to be checked with that file.</param>
        /// <param name="settingsFileOverride">the command line -settings override (if it exists)</param>
        /// <returns>list of violations found in the code.</returns>
        private static List<Violation> ValidateCommittedFiles(Dictionary<string, List<string>> localSourceFiles, string settingsFileOverride)
        {
            string settingsFile = settingsFileOverride;
            if (!string.IsNullOrEmpty(settingsFile))
            {
                if (!Path.IsPathRooted(settingsFile))
                {
                    settingsFile = GetPathInExeFolder(settingsFile);
                }
            }

            // If a settings file has been specified and doesn't exist, it should fall back on the default settings file
            if (string.IsNullOrEmpty(settingsFile) || !File.Exists(settingsFile))
            {
                settingsFile = configSection.StyleCop.SettingsFile;

                if (!Path.IsPathRooted(settingsFile))
                {
                    settingsFile = GetPathInExeFolder(settingsFile);
                }
            }

            List<Violation> violations = new List<Violation>();

            StyleCopConsole console = new StyleCopConsole(settingsFile, false, null, null, true, null);
            List<CodeProject> projects = new List<CodeProject>();
            foreach (KeyValuePair<string, List<string>> config in localSourceFiles)
            {
                StyleCop.Configuration configuration = new StyleCop.Configuration(new string[0]);
                CodeProject project = new CodeProject(string.Empty.GetHashCode(), string.Empty, configuration);

                if (!string.IsNullOrEmpty(config.Key))
                {
                    project.Settings = console.Core.Environment.GetSettings(config.Key, true);
                }

                projects.Add(project);

                foreach (string s in config.Value)
                {
                    console.Core.Environment.AddSourceCode(project, s, null);
                }
            }

            console.ViolationEncountered += (sender, e) => violations.Add(e.Violation);
            console.Start(projects, true);

            return violations;
        }

        /// <summary>
        /// Delete the temporary folder.
        /// </summary>
        private static void DeleteTempFoler()
        {
            string tempPath = GetTempFolderName();
            if (Directory.Exists(tempPath))
            {
                try
                {
                    // Delete the temp directory and all associated files
                    // Catch and swallow all documented thrown exceptions from this in order to
                    // prevent disturbing the committer with a clean up task failure
                    Directory.Delete(tempPath, true);
                }
                catch (IOException)
                {
                }
                catch (UnauthorizedAccessException)
                {
                }
                catch (ArgumentException)
                {
                }
            }
        }

        /// <summary>
        /// Get the full path of a file in the folder of this application.
        /// </summary>
        /// <param name="fileName">name of the file.</param>
        /// <returns>the full path of the file.</returns>
        private static string GetPathInExeFolder(string fileName)
        {
            return Path.Combine(Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location), fileName);
        }

        /// <summary>
        /// Get the temporary folder for this instance.
        /// </summary>
        /// <returns>the full path of the folder.</returns>
        private static string GetTempFolderName()
        {
            string tempRoot = Path.IsPathRooted(configSection.TempFolder) ? configSection.TempFolder : GetPathInExeFolder(configSection.TempFolder);
            return Path.Combine(tempRoot, instanceId.ToString());
        }

        /// <summary>
        /// Print the usage.
        /// </summary>
        private static void PrintUsage()
        {
            Console.WriteLine("Utility to analyze source code files from SVN using StyleCop. Version {0}", Common.ApplicationVersion);
            Console.WriteLine("Copyright (c) Binecs 2009 http://www.binecs.com/");
            Console.WriteLine();
            Console.WriteLine("Usage: {0} [switches]", Path.GetFileName(Assembly.GetExecutingAssembly().Location));
            Console.WriteLine("   -repository:    Path to SVN repository");
            Console.WriteLine("     -revision:    SVN revision number");
            Console.WriteLine("  -transaction:    SVN transaction name");
            Console.WriteLine("     -settings:	  SVN settings file to override default (optional; falls back to default file if it does not exist)");
            Console.WriteLine();
            Console.ForegroundColor = ConsoleColor.Red;
            Console.WriteLine("ERROR: repository and one of transaction or revision must be specified");
            Console.ResetColor();
        }
    }
}
